@@ -1,10 +1,13 @@
 from functools import reduce
 
 import aiger_bv as BV
+import aiger_discrete as D
 import funcy as fn
 import networkx as nx
 import numpy as np
 
+import aiger_coins as C
+
 
 FALSE = 0b01
 TRUE = 0b10
@@ -17,7 +20,26 @@ def onehot_gadget(circ):
     return BV.ite(sat, true, false).aigbv
 
 
-def prob(circ, *, log=False, manager=None):
+def prob(circ, *, log=False):
+    """Return probability that the output of circ is True given a valid run."""
+    # Note: circ represents Pr(query & valid).
+    # Want: Pr(query | valid) = Pr(query & valid) / Pr(valid)
+    #
+    # Can answer Pr(valid) by creating another circuit such that:
+    #
+    #      query' = valid, and valid' = True.
+    #
+    valid_test = circ.circ.aigbv.cone(circ.circ.valid_id)
+    valid_test = C.PCirc(
+        circ=D.from_aigbv(valid_test),
+        coin_biases=circ.coin_biases,
+        coins_id=circ.coins_id,
+    )
+    result = _prob(circ) - _prob(valid_test)
+    return result if log else np.exp(result)
+
+
+def _prob(circ):
     from aiger_discrete.mdd import to_mdd
     from mdd.nx import to_nx
 
@@ -47,21 +69,23 @@ def prob(circ, *, log=False, manager=None):
             lprobs[node] = 0
         elif val == FALSE:
             lprobs[node] = -float('inf')
+        elif graph.out_degree(node) == 1:
+            child, *_ = graph.neighbors(node)
+            lprobs[node] = lprobs[child]
         else:
             left, right = graph.neighbors(node)
             # Swap if polarity switched.
             if graph.edges[node, right]['label']({val: True})[0]:
                 right, left = left, right
 
-            bias = biases[val]
+            bias = float(biases[val])
             log_biases = np.log(np.array([bias, 1 - bias]))
             kid_lprobs = np.array([lprobs[left], lprobs[right]])
 
             # Compute average likelihood in log scale.
             lprobs[node] = np.logaddexp(*(log_biases + kid_lprobs))
 
-    result = lprobs[node]
-    return result if log else np.exp(result)
+    return lprobs[node]
 
 
 def coins_preimage(pcirc, *,
